import unittest
import pinetree as pt


class TestMobileElementManagerMethods(unittest.TestCase):

    def test_init(self):
        manager = pt.MobileElementManager([])

    def test_insert(self):
        pass

    def test_delete(self):
        pass

    def test_choose(self):
        pass

    def test_valid_index(self):
        pass

    def test_get_pol(self):
        pass

    def test_get_attached(self):
        pass

    def test_update_propensity(self):
        pass
